/*
"The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is the content of this file.

The Initial Developer of the Original Code is Dr. Robert Kofler.
Carrer del Baluard; 73 21a
08003 Barcelona
All Rights Reserved.

Contributor(s): Dr. Robert Kofler.

Alternatively, the contents of this file may be used under the terms
of the _____ license (the  "[___] License"), in which case the
provisions of [______] License are applicable instead of those
above. If you wish to allow use of your version of this file only
under the terms of the [____] License and not to allow others to use
your version of this file under the MPL, indicate your decision by
deleting the provisions above and replace them with the notice and
other provisions required by the [___] License. If you do not delete
the provisions above, a recipient may use your version of this file
under either the MPL or the [___] License."
*/
using System;
using System.Collections.Generic;
using System.Text;
using Bio.Repetitive.Microsat;
using Bio.Seq;
using System.IO;
using Bio.IO.Seq;
using Bio.IO.Repetitive.Microsat;
using System.Diagnostics;

namespace SciRoKoSearchSupport
{
    /// <summary>
    /// Provides batch SSR-search capabilities
    /// using the "Factory Method" pattern
    /// </summary>
    public abstract class SSRSearch
    {

        private string ssrOutputFile;
        private string[] sequenceFiles;

        /// <summary>
        /// Initializes SSR search for one file
        /// </summary>
        public SSRSearch(string sequenceFile, string ssrOutputFile)
        {
            this.sequenceFiles = new string[] { sequenceFile };
            this.ssrOutputFile = ssrOutputFile;
        }

        /// <summary>
        /// Initializes SSR search for many files
        /// </summary>
        public SSRSearch(string[] sequenceFiles, string ssrOutputFile)
        {
            this.sequenceFiles = sequenceFiles;
            this.ssrOutputFile = ssrOutputFile;
        }

        public void StartSearch()
        {
            OverallInfo searchInfo = new OverallInfo();
            List<NucleotideSequence> sequences;
            List<SSR> foundSSR = new List<SSR>();

            StreamWriter sw = new StreamWriter(this.ssrOutputFile, false, Encoding.ASCII);
            ISSRWriter ssrwr = SSRUtility.SSRWriterFactory(SSRUtility.GetSSRFileModeFromExtension(Path.GetExtension(this.ssrOutputFile)), sw);

            Debug.WriteLine("");
            Debug.WriteLine(String.Format("{0} SSR-search; SSRWriter created: {1}", DateTime.Now.TimeOfDay, ssrwr.ToString()));
            Debug.WriteLine(String.Format("{0} SSR-search; SSR-searcher used: {1}", DateTime.Now.TimeOfDay, GetSSRSearcherForNucleotideSequence(new NucleotideSequence("fritzi", new Sequence_String("AAA"))).ToString()));


            //Foreach file containing sequences
            foreach (string seqFile in sequenceFiles)
            {

                StreamReader str = new StreamReader(seqFile);
                ISequenceReader<NucleotideSequence> sr = new FastaReader_Sequence<NucleotideSequence>(str, new DNASequenceValidator_ATCGNYRWSKMBDHV(), false);

#if DEBUG
                Stopwatch stow = new Stopwatch();
                stow.Start();
#endif

                //Get a couple of sequences (500)
                while ((sequences = sr.GetSequences()).Count > 0)
                {
                    searchInfo.fileCount += sequences.Count;


#if DEBUG
                    stow.Stop();
                    Debug.WriteLine(String.Format("{0} SSR-search; sequence files have been loaded: {1}; first contained {2} nucleotides; Required time: {3} ms", DateTime.Now.TimeOfDay, sequences.Count, sequences.Count > 0 ? sequences[0].Length : 0, stow.ElapsedMilliseconds));
                    stow.Start();

#endif
                    foreach (INucleotideSequence seq in sequences)
                    {
                        searchInfo.nucleotideCount += seq.Length;
                        ISSRSearcher searcher = GetSSRSearcherForNucleotideSequence(seq);
                        foundSSR.AddRange(searcher.GetSSR());
                    }
#if DEBUG
                    stow.Stop();
                    Debug.WriteLine(String.Format("{0} SSR-search; {1} SSRs have been discovered; Required time: {4} ms", DateTime.Now.TimeOfDay, foundSSR.Count, searchInfo.fileCount, searchInfo.nucleotideCount, stow.ElapsedMilliseconds));
                    stow.Reset();
                    stow.Start();

#endif
                    //Write the found SSRs into the File
                    ssrwr.WriteSSRs(foundSSR);

#if DEBUG
                    stow.Stop();
                    Debug.WriteLine(String.Format("{0} SSR-search; {1} files have been writen in the output file in {2} ms", DateTime.Now.TimeOfDay, foundSSR.Count, stow.ElapsedMilliseconds));
                    Debug.WriteLine("");

#endif

                    foundSSR.Clear();
                    sequences.Clear();

                    System.GC.Collect();

#if DEBUG
                    stow.Reset(); stow.Start(); //For the reading of the files
#endif

                }


                str.Close();
            }
            Debug.WriteLine(String.Format("{0} SSR-search: search has been finished, all ssrs have been reported into output files", DateTime.Now.TimeOfDay));
            //At the end append the Info 
            ssrwr.WriteInfo(searchInfo);
            Debug.WriteLine(String.Format("{0} SSR-search: search info has been writen into the output file; {1} fastas files haven been searched; number of nucleotides {2}", DateTime.Now.TimeOfDay, searchInfo.fileCount, searchInfo.nucleotideCount));
            Debug.WriteLine("");
            sw.Close();
            System.Diagnostics.Debug.Flush();
        }

        protected abstract ISSRSearcher GetSSRSearcherForNucleotideSequence(INucleotideSequence sequence);



    }




    public class Model_SSRSearch
    {
        private TimeSpan searchTime;
        private string[] sequenceFilesForAnalysis;
        private string saveFile;

        public Model_SSRSearch()
        {
        }

        public void Reset()
        {
            this.searchTime = new TimeSpan();
            this.sequenceFilesForAnalysis = null;
            this.saveFile = null;
        }

        /// <summary>
        /// Sequence files (fasta) in which to search for SSRs 
        /// </summary>
        public string[] SequenceFiles
        {
            get
            {
                return sequenceFilesForAnalysis;
            }
            set
            {
                this.sequenceFilesForAnalysis = value;
                //Debug code does not affect the Release
                Debug.WriteLine(String.Format("{0} SciRoKo SSR-search model; {1} sequence files have been chosen for analysis", DateTime.Now.TimeOfDay, value.Length));
            }



        }


        /// <summary>
        /// The file chosen for saving the results of the SSR-search
        /// </summary>
        public string SaveFile
        {
            get
            {
                return saveFile;
            }
            set
            {
                saveFile = value;
                //Debug code does not affect the performance of the release
                Debug.WriteLine(String.Format("{0} SciRoKo SSR-search model; Output file has been chosen: {1}", DateTime.Now.TimeOfDay, value));
            }
        }

        /// <summary>
        /// The required SSR-search time
        /// </summary>
        public TimeSpan SearchTime
        {
            get
            {
                return searchTime;
            }
        }


        public void StartSSRSearch(SSRSearchSettings searchSettings)
        {



            SSRSearcherSciRoKo searcher = new SSRSearcherSciRoKo(searchSettings, this.sequenceFilesForAnalysis, this.saveFile);


            Stopwatch stopW = new Stopwatch();
            stopW.Start();

            Debug.WriteLine(String.Format("{0} SciRoKo SSR-search model; SSR-search has been started", DateTime.Now.TimeOfDay));
            searcher.StartSearch();
            stopW.Stop();
            Debug.WriteLine(String.Format("{0} SciRoKo SSR-search model; SSR-search has been finished in {1} ms", DateTime.Now.TimeOfDay, stopW.ElapsedMilliseconds));
            Debug.WriteLine("");





            this.searchTime = stopW.Elapsed;




        }


    }

    /// <summary>
    /// The special implementation for SSR search used in SciRoKo;
    /// The factory method has to be implemented
    /// </summary>
    public sealed class SSRSearcherSciRoKo : SSRSearch
    {
        SSRSearchSettings settings;
        public SSRSearcherSciRoKo(SSRSearchSettings settings, string[] sequenceFiles, string ssrOutput)
            : base(sequenceFiles, ssrOutput)
        {
            this.settings = settings;
        }

        /// <summary>
        /// Factory Method, implemented in the derived class, especially to fit the requirements of SciRoKo.
        /// </summary>
        protected override ISSRSearcher GetSSRSearcherForNucleotideSequence(INucleotideSequence sequence)
        {
            //Switch all possible search modes
            switch (settings.searchMode)
            {
                case SSRSearchMode.PerfectRepeats: return new PerfectSSRSearch_MinimumRepeats(sequence, 3, settings.minRep);
                case SSRSearchMode.PerfectLength: return new PerfectSSRSearch_TotalLength(sequence, settings.minLength, settings.minRep);
                case SSRSearchMode.PerfectMISA: return new PerfectSSRSearch_MISA(sequence, settings.misaDefinition);
                case SSRSearchMode.MisMatchFixedPenalty: return new MismatchedSSRSearch_FixedPenalty(sequence, settings.requiredScore, settings.seedMinLength, settings.seedMinRep, settings.mmao, settings.mismatchPenalty);
                case SSRSearchMode.MisMatchVariablePenalty: return new MismatchedSSRSearch_VariablePenalty(sequence, settings.requiredScore, settings.seedMinLength, settings.seedMinRep, settings.mmao, settings.mismatchPenalty);
                default: throw new Exception("Error during SSR search; no valid SSRSearch mode has been specified");

            }

        }

    }



    public struct SSRSearchSettings
    {

        public SSRSearchMode searchMode;
        public string misaDefinition;
        public int minRep;
        public int minLength;
        public int mismatchPenalty; //Variable or fixed
        public int mmao;
        public int seedMinLength;
        public int seedMinRep;
        public int requiredScore;

    }


}
