/*
"The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is the content of this file.

The Initial Developer of the Original Code is Dr. Robert Kofler.
Carrer del Baluard; 73 21a
08003 Barcelona
All Rights Reserved.

Contributor(s): Dr. Robert Kofler.

Alternatively, the contents of this file may be used under the terms
of the _____ license (the  "[___] License"), in which case the
provisions of [______] License are applicable instead of those
above. If you wish to allow use of your version of this file only
under the terms of the [____] License and not to allow others to use
your version of this file under the MPL, indicate your decision by
deleting the provisions above and replace them with the notice and
other provisions required by the [___] License. If you do not delete
the provisions above, a recipient may use your version of this file
under either the MPL or the [___] License."
*/
using System;
using System.Collections.Generic;
using System.Windows.Forms;
using System.Xml;
using System.Text;
using System.IO;
using System.Diagnostics;
using Bio.IO.Repetitive.Microsat;
using Bio.IO.Seq;
using Bio.Seq;
using Bio.Repetitive.Microsat;
using Bio.Repetitive.Microsat.Statistics;
using SciRoKoSearchSupport;

namespace SciRoKo
{
    /// <summary>
    /// The controll updating the SciRoKo elements and Views.
    /// </summary>
    public class SciRoKoControl
    {
        SciRoKo view;
        ExtractionControl flank;
        Model_SSRSearch mSearch;
        Model_SSRStatistic mStatistic;
        Model_IniFileAccess mIni;
        Model_LittleHelper mHelper;
        ISciRoKoMain_State activeState;
        public SciRoKoControl(SciRoKo view)
        {
            this.view = view;
            this.mIni = new Model_IniFileAccess();
            this.mSearch = new Model_SSRSearch();
            this.mStatistic = new Model_SSRStatistic();
            this.mHelper = new Model_LittleHelper();
            activeState = new SciRoKoState_None(view);
            this.SearchModeChanged();
            view.LabelStartupPathText = this.StartupPath;
        }

        public string StartupPath
        {
            get
            {
                return mIni.StartupPath;
            }
        }

        /// <summary>
        /// The Reset Button has been clicked
        /// </summary>
        public void ButtonResetClick()
        {
            Debug.WriteLine(String.Format("{0} SciRoKo-control; Reset button hit", DateTime.Now.TimeOfDay));
            //Always updated the model first, secondly the view.
            mSearch.Reset();
            mStatistic.Reset();

            this.activeState = new SciRoKoState_None(view);


        }

        /// <summary>
        /// The load Sequences button has been clicked
        /// </summary>
        public void ButtonLoadSequencesClick(string[] sequenceFilesForAnalysis)
        {
            mStatistic.Reset();
            Debug.WriteLine(String.Format("{0} SciRoKo-control; Load sequences button hit", DateTime.Now.TimeOfDay));
            try
            {

                mSearch.SequenceFiles = sequenceFilesForAnalysis;
                if (this.activeState is SciRoKoState_SaveFileChosen)
                {
                    this.activeState = new SciRoKoState_SequencesAndSaveFilesChosen(view, mSearch);
                }
                else
                {
                    this.activeState = new SciRoKoState_SequencesChosen(view, mSearch);
                }
            }
            catch (Exception e)
            {
                MessageBox.Show("Loading Sequences not possible: " + e.Message);
                this.ButtonResetClick();
            }
        }

        public void ButtonOutputFileClick(string outputFile)
        {
            mStatistic.Reset();
            Debug.WriteLine(String.Format("{0} SciRoKo-control; Select output file button hit", DateTime.Now.TimeOfDay));
            try
            {
                mSearch.SaveFile = outputFile;
                if (this.activeState is SciRoKoState_SequencesChosen || this.activeState is SciRoKoState_SearchFinished)
                {
                    this.activeState = new SciRoKoState_SequencesAndSaveFilesChosen(view, mSearch);
                }
                else
                {
                    this.activeState = new SciRoKoState_SaveFileChosen(view, mSearch);
                }
            }
            catch (Exception e)
            {
                MessageBox.Show("Access to output file denied: " + e.Message);
                this.ButtonResetClick();
            }
        }

        /// <summary>
        /// 
        /// </summary>
        public void ButtonStartSearchClick(SSRSearchSettings searchSettings)
        {
            Debug.WriteLine(String.Format("{0} SciRoKo-control; Start SSR-search button hit", DateTime.Now.TimeOfDay));
            Debug.Assert((this.activeState is SciRoKoState_SequencesAndSaveFilesChosen), "Controll state is not valid");


            mSearch.StartSSRSearch(searchSettings);

            if (!view.CheckBox_LoadSearchIntoStatisticCheck)
            {
                mSearch.SaveFile = null;
                activeState = new SciRoKoState_SearchFinished(view, mSearch);

            }
            else
            {
                this.ButtonLoadSSRsClick(new string[] { mSearch.SaveFile });

            }
        }

        public void ButtonLoadSSRsClick(string[] microsatelliteFilesForAnalysis)
        {
            Debug.WriteLine(String.Format("{0} SciRoKo-control; Load SSRs button hit", DateTime.Now.TimeOfDay));

            mStatistic = new Model_SSRStatistic();
            string extension = Path.GetExtension(microsatelliteFilesForAnalysis[0]);
            if (extension == ".ssrCou") //if the SSR couple mode has been chosen
            {

                mStatistic.LoadSSRCouples(microsatelliteFilesForAnalysis);
                activeState = new SciRoKoState_SSRCouples(view, mStatistic);
            }
            else
            {
                mStatistic.LoadSSRs(microsatelliteFilesForAnalysis);
                mStatistic.SetSSRSubset(view.SubsetBoundaries);
                activeState = new SciRoKoState_SSRLoaded(view, mStatistic);
            }
            this.SubsetChanged();

        }

        public void ButtonExportSSRsClick(string exportFile)
        {
            Debug.WriteLine(String.Format("{0} SciRoKo-control; Export SSRs button hit", DateTime.Now.TimeOfDay));

            string extension = Path.GetExtension(exportFile);
            if (extension == ".ssrCou") //if the SSR couple mode has been chosen
            {
                if (activeState is SciRoKoState_SSRCouples) mStatistic.ExportSSRCouples(exportFile, view.StatisticParameters, false);
                else if (activeState is SciRoKoState_SSRLoaded) mStatistic.ExportSSRCouples(exportFile, view.StatisticParameters, true);
                else throw new Exception("SciRoKo controll: invalid state for exporting microsatellites");
            }
            else
            {
                mStatistic.ExportSSRs(exportFile, view.StatisticParameters);
            }

        }

        public void ButtonDisplayStatisticClick()
        {
            Debug.WriteLine(String.Format("{0} SciRoKo-control; Display SSRs-statistics button hit", DateTime.Now.TimeOfDay));

            StringBuilder sb;
            if (!view.RadioButton_DisplayStatitsticsChecked)
            {
                sb = mStatistic.DisplaySSRSearchResults(view.StatisticParameters);

            }
            else
            {
                sb = mStatistic.DisplaySSRStatistics(view.StatisticParameters, (activeState is SciRoKoState_SSRCouples));

            }


            view.RichTextBox_Analyzed_StringBuilder = sb;
        }

        public void SearchModeChanged()
        {
            Debug.WriteLine(String.Format("{0} SciRoKo-control; SSR-search mode changed to: {1}", DateTime.Now.TimeOfDay, view.SelectedSSRSearchMode.ToString()));


            switch (view.SelectedSSRSearchMode)
            {
                case SSRSearchMode.PerfectRepeats:
                    view.Label_RequirementPerfectText = "Minimum repeats:";
                    view.TextBox_RequirementPerfectEnabled = true;
                    view.TextBox_RequirementPerfectText = "4";
                    view.TextBox_MinRepeatsEnabled = false;
                    view.TextBox_MismatchPenaltyEnabled = false;
                    view.TextBox_RequiredScoreEnabled = false;
                    view.TextBox_SSRSeedLengthEnabled = false;
                    view.TextBox_SSRSeedRepeatsEnabled = false;
                    view.TextBox_MaxMismatchesAtOnceEnabled = false;
                    break;
                case SSRSearchMode.PerfectLength:
                    view.Label_RequirementPerfectText = "Minimum total length:";
                    view.TextBox_RequirementPerfectEnabled = true;
                    view.TextBox_RequirementPerfectText = "15";
                    view.TextBox_MinRepeatsEnabled = true;
                    view.TextBox_MismatchPenaltyEnabled = false;
                    view.TextBox_RequiredScoreEnabled = false;
                    view.TextBox_SSRSeedLengthEnabled = false;
                    view.TextBox_SSRSeedRepeatsEnabled = false;
                    view.TextBox_MaxMismatchesAtOnceEnabled = false;
                    break;
                case SSRSearchMode.PerfectMISA:
                    view.Label_RequirementPerfectText = "Min. rep. (eg.:6-6-5-4-3-3):";
                    view.TextBox_RequirementPerfectEnabled = true;
                    view.TextBox_RequirementPerfectText = "14-7-5-4-4-4";
                    view.TextBox_MinRepeatsEnabled = false;
                    view.TextBox_MismatchPenaltyEnabled = false;
                    view.TextBox_RequiredScoreEnabled = false;
                    view.TextBox_SSRSeedLengthEnabled = false;
                    view.TextBox_SSRSeedRepeatsEnabled = false;
                    view.TextBox_MaxMismatchesAtOnceEnabled = false;
                    break;
                case SSRSearchMode.MisMatchFixedPenalty:
                    view.TextBox_RequirementPerfectEnabled = false;
                    view.TextBox_MinRepeatsEnabled = false;
                    view.TextBox_MismatchPenaltyEnabled = true;
                    view.TextBox_RequiredScoreEnabled = true;
                    view.TextBox_SSRSeedLengthEnabled = true;
                    view.TextBox_SSRSeedRepeatsEnabled = true;
                    view.TextBox_MaxMismatchesAtOnceEnabled = true;
                    view.Label_MismatchPenaltyText = "Mismatch penalty:";
                    view.TextBox_MismatchPenaltyText = "5";
                    break;
                case SSRSearchMode.MisMatchVariablePenalty:
                    view.TextBox_RequirementPerfectEnabled = false;
                    view.TextBox_MinRepeatsEnabled = false;
                    view.TextBox_MismatchPenaltyEnabled = true;
                    view.TextBox_RequiredScoreEnabled = true;
                    view.TextBox_SSRSeedLengthEnabled = true;
                    view.TextBox_SSRSeedRepeatsEnabled = true;
                    view.TextBox_MaxMismatchesAtOnceEnabled = true;
                    view.Label_MismatchPenaltyText = "Mismatch penalty: (Motif length * X)";
                    view.TextBox_MismatchPenaltyText = "1";
                    break;
            }
        }

        public void SubsetChanged()
        {
            Debug.WriteLine(String.Format("{0} SciRoKo-control; SSR-subset boundarys changed", DateTime.Now.TimeOfDay));
            mStatistic.SetSSRSubset(view.SubsetBoundaries);
            activeState.UpdateView();
        }

        public void Button_PathSelectedClick(string path)
        {
            Debug.WriteLine(String.Format("{0} SciRoKo-control; A new startup path has been selected: {1}", DateTime.Now.TimeOfDay, path));
            mIni.StartupPath = path;
            view.LabelStartupPathText = mIni.StartupPath;

        }

        public void Button_DigestSequencesClick(string[] sequenceFiles, DigestionSettings settings)
        {
            Debug.WriteLine(String.Format("{0} SciRoKo-control; The sequence digestion subunit has been activated", DateTime.Now.TimeOfDay));
            mHelper.DigestSequences(sequenceFiles, settings);
        }

        public void Button_ExtractFlankingSequencesClick()
        {

            if (this.activeState is SciRoKoState_SSRLoaded)
            {
                flank = new ExtractionControl(this, mStatistic.GetSSRSubset());
            }
            else if (this.activeState is SciRoKoState_SSRCouples)
            {
                flank = new ExtractionControl(this, mStatistic.GetSSRCouples());
            }
            else throw new Exception("SciRoKo state is not appropriate for extracting flanking sequences");


            


        }



    }

    public class ExtractionControl
    {


        private Model_LittleHelper mHelper;
        private ExtractFlank flankDialog;
        private List<IPositionable> pos;
        private SciRoKoControl control;
        private string text = "";
        private int size = 0;
        


        public ExtractionControl(SciRoKoControl control, List<SSR> ssrs)
        {
            this.control = control;

            text = "SSRs";
            this.mHelper = new Model_LittleHelper();
            this.mHelper.SequencesToFlank = ssrs.ConvertAll<IPositionable>(new Converter<SSR, IPositionable>(delegate(SSR ssr) { return (IPositionable)ssr; }));
            size = ssrs.Count;
            this.Initialze();
        }
        public ExtractionControl(SciRoKoControl control, List<SSRCouple> couples)
        {
            this.control = control;
            text = "SSR-Couples";
            this.mHelper = new Model_LittleHelper();
            this.mHelper.SequencesToFlank = couples.ConvertAll<IPositionable>(new Converter<SSRCouple, IPositionable>(delegate(SSRCouple couple) { return (IPositionable)couple; }));
            size = couples.Count;
            this.Initialze();
        }


        public void Initialze()
        {
            flankDialog = new ExtractFlank(this);
            flankDialog.LabelExtractText_Text=String.Format("{0} {1} are pending extraction. Chose the file containing the nucleotide sequences for which the SSR-search has been conducted",size,text);
            flankDialog.ShowDialog();

            
        }
        public string StartupPath
        {
            get
            {
                return control.StartupPath;
            }
        }

        public void ButtonChoseInputFileClick(string[] inputFiles)
        {
            flankDialog.ButtonStartExtraction_Enabled = false;
            this.mHelper.FilesContainingTheFlanks = inputFiles;
            StringBuilder sb = new StringBuilder("");

            foreach(string s in inputFiles)
            {
                sb.Append(Path.GetFileName(s) +'\n');
            }
            flankDialog.TextBoxChoseInputFile_Text = sb.ToString();


            flankDialog.ButtonChoseOutputFile_Enabled = true;
            flankDialog.ButtonStartExtraction_Enabled = false;
        }

        public void ButtonChoseOutputFileClick(string outputFile)
        {
            flankDialog.ButtonStartExtraction_Enabled = true;
            this.mHelper.SaveFileForFlankingSequences = outputFile;
            StringBuilder sb = new StringBuilder("");
            sb.Append(Path.GetFileName(outputFile));
            
            flankDialog.TextBoxChoseOutputFile_Text = sb.ToString();

            flankDialog.ButtonChoseOutputFile_Enabled = true;
            flankDialog.ButtonStartExtraction_Enabled = true;
        }

        public void ButtonStartExtractionClick(ExtractFlankingSettings setting)
        {
            mHelper.ExtractFlankingSequences(setting);
            MessageBox.Show("Succesfully extracted flanking sequences for microsatellites");


            flankDialog.ButtonChoseOutputFile_Enabled = false;
            flankDialog.ButtonStartExtraction_Enabled = false;
            flankDialog.TextBoxChoseInputFile_Text = "";
            flankDialog.TextBoxChoseOutputFile_Text = "";
            mHelper = new Model_LittleHelper();
        }

        public void FlankingModeChanged(bool variableFlankMode)
        {
            if (variableFlankMode)
            {
                flankDialog.Label5PrimeFlank_Visble = true;
                flankDialog.Label5PrimeFank_Text = "5' - flanking";
                flankDialog.TextBox5PrimeFlank_Text = "200";


                flankDialog.Label3PrimeFlank_Visible = true;
                flankDialog.TextBox3PrimeFlank_Visible = true;
                flankDialog.TextBox3PrimeFlank_Text = "200";

            }
            else
            {
                flankDialog.Label5PrimeFlank_Visble = true;
                flankDialog.Label5PrimeFank_Text = "Total flanking";
                flankDialog.TextBox5PrimeFlank_Text = "400";


                flankDialog.Label3PrimeFlank_Visible = false;
                flankDialog.TextBox3PrimeFlank_Visible = false;
            }
        }
    }


  


    public class Model_SSRStatistic
    {
        private string[] microsatelliteFilesForStatistics;
        private OverallInfo nucleotideSummaryInfo;
        private SSRSubsetBoundary subsetBoundary;
        private List<SSRCouple> totalCouples;
        private List<SSR> totalSSRs;
        private List<SSR> subsetSSRs;

        public Model_SSRStatistic()
        {
            this.Reset();
        }


        public void Reset()
        {
            this.nucleotideSummaryInfo.fileCount = 0;
            this.nucleotideSummaryInfo.nucleotideCount = 0;
            this.microsatelliteFilesForStatistics = null;
            subsetBoundary = new SSRSubsetBoundary();
            this.totalCouples = new List<SSRCouple>();
            this.totalSSRs = new List<SSR>();
            this.subsetSSRs = new List<SSR>();
        }

        /// <summary>
        /// all microsatellite files chosen for the analysis, and statistical interpretation
        /// </summary>
        public string[] MicrosatelliteFiles
        {
            get
            {
                return this.microsatelliteFilesForStatistics;
            }
        }


        /// <summary>
        /// Calculates the Subset of microsatellites which should be used for the analysis
        /// </summary>
        private void CalculateSubset()
        {
            if (this.totalSSRs != null)
            {
                this.subsetSSRs = SSRUtility.GetSubsetScore(totalSSRs, subsetBoundary.minimumScore);
                this.subsetSSRs = SSRUtility.GetSubsetMotifLength(subsetSSRs, subsetBoundary.lowerBoundary, subsetBoundary.upperBoundary);
            }
            Debug.WriteLine(String.Format("{0} SciRoKo SSR-statistics model; A new SSR-subset has been created containing {1} SSRs", DateTime.Now.TimeOfDay, subsetSSRs.Count));
        }


        /// <summary>
        /// Summary information about the fasta sequences with which the SSR-search has been conducted
        /// </summary>
        public OverallInfo SequenceSummaryInfo
        {
            get { return nucleotideSummaryInfo; }
            set { nucleotideSummaryInfo = value; }
        }


        /// <summary>
        /// Returns the number of SSRs contained in the selected subset
        /// </summary>
        public int SSRSubsetCount
        {
            get
            {
                return subsetSSRs.Count;
            }
        }

        /// <summary>
        /// Returns the total number of microsatellites
        /// </summary>
        public int SSRTotalCount
        {
            get
            {
                return totalSSRs.Count;
            }
        }

        /// <summary>
        /// Returns the total number of SSR-Couples
        /// </summary>
        public int SSRCoupleCount
        {
            get
            {
                return totalCouples.Count;
            }
        }


        /// <summary>
        /// Sets the boundary for the SSR-subset
        /// </summary>
        public void SetSSRSubset(SSRSubsetBoundary subsetBoundary)
        {


            this.subsetBoundary = subsetBoundary;
            CalculateSubset();

        }


        public List<SSR> GetSSRSubset()
        {
            return subsetSSRs;
        }

        public List<SSRCouple> GetSSRCouples()
        {
            return totalCouples;
        }

        /// <summary>
        /// Load the microsatellites into the statistic module
        /// The subset is not calculated
        /// </summary>
        public void LoadSSRs(string[] microsatelliteFilesForAnalysis)
        {
            Reset();
            this.microsatelliteFilesForStatistics = microsatelliteFilesForAnalysis;
            SSRFileMode ssrFileMode = SSRUtility.GetSSRFileModeFromExtension(Path.GetExtension(microsatelliteFilesForStatistics[0]));
#if DEBUG
            Stopwatch stopw = new Stopwatch();
            stopw.Start();
#endif
            foreach (string s in microsatelliteFilesForStatistics)
            {
                List<SSR> temp = new List<SSR>();
                StreamReader sr = new StreamReader(s, Encoding.ASCII);
                ISSRReader ssrr = SSRUtility.SSRReaderFactory(ssrFileMode, sr);
                try
                {




                    while ((temp = ssrr.ReadSSRs()).Count > 0)
                    {
                        totalSSRs.AddRange(temp);
                        temp.Clear();
                    }
                    OverallInfo info = ssrr.ReadInfo();

                    nucleotideSummaryInfo.nucleotideCount += info.nucleotideCount;
                    nucleotideSummaryInfo.fileCount += info.fileCount;
                }
                finally
                {
                    //Close the streamreader
                    sr.Close();
                }
            }

#if DEBUG
            stopw.Stop();
            Debug.WriteLine(String.Format("{0} SciRoKo SSR-statistics model; {1} SSRs have been loaded from {2} files in {3} ms", DateTime.Now.TimeOfDay, this.totalSSRs.Count, microsatelliteFilesForAnalysis.Length, stopw.ElapsedMilliseconds));
            Debug.WriteLine("");
#endif



        }

        /// <summary>
        /// Load the SSR-Couples into the statistic module the subset is not calculated
        /// </summary>
        public void LoadSSRCouples(string[] microsatelliteFilesForAnalysis)
        {

            this.totalCouples = new List<SSRCouple>();
            this.microsatelliteFilesForStatistics = microsatelliteFilesForAnalysis;
            this.nucleotideSummaryInfo.fileCount = 0;
            this.nucleotideSummaryInfo.nucleotideCount = 0;
            this.subsetSSRs = null;
            this.totalSSRs = null;
#if DEBUG
            Stopwatch stopw = new Stopwatch();
            stopw.Start();
#endif
            foreach (string s in microsatelliteFilesForStatistics)
            {
                StreamReader sr = new StreamReader(s, Encoding.ASCII);
                SSRCoupleReader cr = new SSRCoupleReader(sr);
                try
                {

                    List<SSRCouple> temp;
                    while ((temp = cr.ReadCouples()).Count > 0)
                    {
                        totalCouples.AddRange(temp);
                        temp.Clear();
                    }
                    this.nucleotideSummaryInfo.fileCount += cr.ReadInfo().fileCount;
                    this.nucleotideSummaryInfo.nucleotideCount += cr.ReadInfo().nucleotideCount;
                }
                finally
                {
                    sr.Close();
                }
            }

#if DEBUG
            stopw.Stop();
            Debug.WriteLine(String.Format("{0} SciRoKo SSR-statistics model; {1} SSR-couples have been loaded from {2} files in {3} ms", DateTime.Now.TimeOfDay, this.totalCouples.Count,this.microsatelliteFilesForStatistics.Length, stopw.ElapsedMilliseconds));
            Debug.WriteLine("");
#endif

        }



        public void ExportSSRs(string exportFile, SSRStatisticParameters parameters)
        {
            if (subsetSSRs == null) throw new Exception("SciRoKo SSR-statistics module: failure during exporting of the microsatellites, subset has not been created");
            SSR.SSRSort sorter = new SSR.SSRSort(parameters.ssrSortMode);
            subsetSSRs.Sort(sorter);

            StreamWriter sw = new StreamWriter(exportFile, false, Encoding.ASCII);
#if DEBUG
            Stopwatch stopw = new Stopwatch();
            stopw.Start();
#endif
            try
            {




                ISSRWriter ssrwr = SSRUtility.SSRWriterFactory(SSRUtility.GetSSRFileModeFromExtension(Path.GetExtension(exportFile)), sw);


                ssrwr.WriteSSRs(subsetSSRs);
                ssrwr.WriteInfo(nucleotideSummaryInfo);
            }
            finally
            {
                sw.Close();
            }

#if DEBUG
            stopw.Stop();
            Debug.WriteLine(String.Format("{0} SciRoKo SSR-statistics model; {1} SSRs have been exported into {2} in {3} ms", DateTime.Now.TimeOfDay, this.subsetSSRs.Count, exportFile,stopw.ElapsedMilliseconds));
            Debug.WriteLine("");
#endif

        }

        public void ExportSSRCouples(string exportFile, SSRStatisticParameters parameters, bool calculateCouplesNew)
        {
            List<SSRCouple> toSave;

            //Aktualise the SSR-Couples either use the SSR-couples already in use or obtain new SSR-Couples
            if (calculateCouplesNew) toSave =SSRClusterUtility.Get_SSRCouple(subsetSSRs, parameters.maxDistanceForAssociation);
            else toSave = totalCouples;




            if (parameters.couplesSubtype == SSRStatisticParameters.CoupleSubtype.Heterocouples)
                toSave = SSRClusterUtility.Get_HeteroCouple(toSave);
            else if (parameters.couplesSubtype == SSRStatisticParameters.CoupleSubtype.Homocouples)
                toSave = SSRClusterUtility.Get_HomoCouple(toSave);

            StreamWriter sw = new StreamWriter(exportFile, false, Encoding.ASCII);

#if DEBUG
            Debug.WriteLine(String.Format("{0} SSR-statistics module; SSR-couple export, chosen subset {1}",DateTime.Now.TimeOfDay,parameters.couplesSubtype.ToString()));
            Stopwatch stopw = new Stopwatch();
            stopw.Start();
#endif

            try
            {

                SSRCoupleWriter cw = new SSRCoupleWriter(sw);
                cw.WriteCouples(toSave);
                cw.WriteInfo(nucleotideSummaryInfo);

            }
            finally
            {
                sw.Close();
            }

#if DEBUG
            stopw.Stop();
            Debug.WriteLine(String.Format("{0} SciRoKo SSR-statistics model; {1} SSR-couples have been exported into {2} in {3} ms", DateTime.Now.TimeOfDay, this.subsetSSRs.Count, exportFile, stopw.ElapsedMilliseconds));
            Debug.WriteLine("");
#endif
        }

        public StringBuilder DisplaySSRSearchResults(SSRStatisticParameters parameters)
        {
            if (subsetSSRs == null) throw new Exception("SSR-statistics module: subset is in invalid state (null)");
            StringBuilder sb = new StringBuilder("");
            SSR.SSRSort sorter = new SSR.SSRSort(parameters.ssrSortMode);

#if DEBUG
            Stopwatch stopw = new Stopwatch();
            stopw.Start();
#endif
            subsetSSRs.Sort(sorter);


            //First create the header 
            sb.Append(SSR.ToStringHeader() + '\n');

            foreach (SSR s in subsetSSRs)
            {
                sb.Append(s.ToString(30) + '\n');
            }
#if DEBUG
            stopw.Stop();
            Debug.WriteLine(String.Format("{0} SciRoKo SSR-statistics model; Displayed SSR-search results in {1} ms; {2} SSRs in subset to display", DateTime.Now.TimeOfDay, stopw.ElapsedMilliseconds, subsetSSRs.Count));
            Debug.WriteLine("");
#endif

            //then load the whole string at once into the Rich text box
            return sb;
        }



        public StringBuilder DisplaySSRStatistics(SSRStatisticParameters parameters, bool isInCoupleMode)
        {
            StringBuilder sb = new StringBuilder("");
            Bio.Repetitive.Microsat.Statistics.SSRStatistics_DetailLevel detailLevel = parameters.statisticDetailLevel;
            int minCount = parameters.minimumFrequencyInCount;
#if DEBUG
            Stopwatch stopw=new Stopwatch();
            stopw.Start();
#endif


            if (!isInCoupleMode)
            {
                Debug.WriteLine(String.Format("{0} SSR-statistics module; Display statistics; Not in SSR-couple mode", DateTime.Now.TimeOfDay));
                if (subsetSSRs == null) throw new Exception("SSR-statistics module: subset is in invalid state (null)");

                SSR_StatisticsProvider stat = new SSR_StatisticsProvider(subsetSSRs, nucleotideSummaryInfo);



                List<SSRStatisticMotifLengthInfo> lengthInfos = stat.GetSSRStatisticMotifLengthInfo();

                //First: Calculate some basic information
                int totalMismatches = 0;
                long totalLength = 0;
                foreach (SSR s in subsetSSRs)
                {
                    totalMismatches += s.Mismatches;
                    totalLength += s.Length;

                }
                double averageLength = (double)totalLength / (double)subsetSSRs.Count;


                //Calculate standard deviation
                double squareSumL = 0.0;
                foreach (SSR s in subsetSSRs)
                {
                    squareSumL += Math.Pow((s.Length - averageLength), 2.0);

                }
                double avStdDevL = Math.Sqrt(squareSumL / subsetSSRs.Count);



                //DISPLAY:

                sb.Append(string.Format("SSRs in Subset: {0} - Average mismatches: {1,5:f} - Average length: {2,5:f} - Average Standard deviation length {3,5:f} - counts/Mbp: {4,7:f}\n"
                    , subsetSSRs.Count, (double)totalMismatches / (double)subsetSSRs.Count, averageLength, avStdDevL, 1000000.0 / ((double)nucleotideSummaryInfo.nucleotideCount / (double)subsetSSRs.Count)));
                sb.Append(String.Format("\n\n"));
                Debug.WriteLine(String.Format("{0} SSR-statistics module; Display statistics; Displayed overall statistics", DateTime.Now.TimeOfDay));

                //MotifLength statistics:
                if (parameters.displayMotifLengthStatistics)
                {


                    sb.Append(SSRStatisticMotifLengthInfo.ToStringHeader(detailLevel) + '\n');
                    foreach (SSRStatisticMotifLengthInfo li in lengthInfos)
                    {
                        sb.Append(li.ToString(detailLevel) + '\n');
                    }
                    sb.Append('\n');
                    Debug.WriteLine(String.Format("{0} SSR-statistics module; Display statistics; Displayed motif length statistics", DateTime.Now.TimeOfDay));
                }


                //MOTIF INFOS
                if (parameters.displayMotifStatistics)
                {
                    SSRStatisticMotifInfo.SSRStatisticMotifInfoSort sort = new SSRStatisticMotifInfo.SSRStatisticMotifInfoSort(SSRStatisticMotifInfo.SSRStatisticMotifInfoSortMode.TotalCount);


                    if (parameters.motifStat_FS)
                    {
                        List<SSRStatisticMotifInfo> motifInfos = stat.GetSSRStatisticMotifInfo(SSRStandardization_Degree.Fully);
                        motifInfos.Sort(sort);


                        //MotifInfo:
                        sb.Append("Motif statistics - Fully standardized\n");
                        sb.Append(SSRStatisticMotifInfo.ToStringHeader(detailLevel) + '\n');
                        foreach (SSRStatisticMotifInfo mi in motifInfos)
                        {
                            if (mi.Count >= minCount)
                            {
                                sb.Append(mi.ToString(detailLevel) + '\n');
                            }
                        }
                        sb.Append('\n');
                        Debug.WriteLine(String.Format("{0} SSR-statistics module; Display statistics; Displayed fully standardized motif statistic", DateTime.Now.TimeOfDay));

                    }


                    if (parameters.motifStat_PS)
                    {
                        List<SSRStatisticMotifInfo> motifInfos = stat.GetSSRStatisticMotifInfo(SSRStandardization_Degree.Partially);
                        motifInfos.Sort(sort);

                        //MotifInfo:
                        sb.Append("Motif statistics - Partially standardized\n");
                        sb.Append(SSRStatisticMotifInfo.ToStringHeader(detailLevel) + '\n');
                        foreach (SSRStatisticMotifInfo mi in motifInfos)
                        {
                            if (mi.Count >= minCount)
                            {
                                sb.Append(mi.ToString(detailLevel) + '\n');
                            }
                        }
                        sb.Append('\n');
                        Debug.WriteLine(String.Format("{0} SSR-statistics module; Display statistics; Displayed partially standardized motif statistic", DateTime.Now.TimeOfDay));
                    }
                }
            }

            //MOTIF ASSOCIATIONS
            if (parameters.displayMotifAssociationStatistics)
            {
                List<SSRCouple> all_couples;
                if (!isInCoupleMode)
                {
                    all_couples = SSRClusterUtility.Get_SSRCouple(subsetSSRs, parameters.maxDistanceForAssociation);
                }
                else
                {
                    all_couples = totalCouples;
                }


                List<SSRCouple> activeCouples;
                List<SSRCouple> homoCouples = SSRClusterUtility.Get_HomoCouple(all_couples);
                List<SSRCouple> heteroCouples = SSRClusterUtility.Get_HeteroCouple(all_couples);

                //Check if the determination of the couples have been done correct
                if (all_couples.Count != (homoCouples.Count + heteroCouples.Count)) throw new Exception("Error during determination of the number of Homo-, Heterocouples. Counts do not add up.");

                string motifAssocText = "";
                //Set the active couple 

                if (parameters.couplesSubtype == SSRStatisticParameters.CoupleSubtype.All)
                {
                    activeCouples = all_couples;
                    motifAssocText = "Motif association statistic has been done with all SSR-Couples (Homo- and Heterocouples)\n";
                }
                else if (parameters.couplesSubtype == SSRStatisticParameters.CoupleSubtype.Homocouples)
                {
                    activeCouples = homoCouples;
                    motifAssocText = "Motif association statistic has been done with Homocouples\n";

                }
                else if (parameters.couplesSubtype == SSRStatisticParameters.CoupleSubtype.Heterocouples)
                {
                    activeCouples = heteroCouples;
                    motifAssocText = "Motif association statistic has been done with Heterocouples\n";
                }
                else throw new Exception("Impossible mistake occured");




                //Initialize Couple statistic
                SSRCouple_StatisticsProvider couStat = new SSRCouple_StatisticsProvider(activeCouples, nucleotideSummaryInfo);


                //The general statistics are already done with the active couples.
                int mmCount = 0; //Total mismatches
                int maCount = activeCouples.Count;
                long leCount = 0;
                foreach (SSRCouple c in activeCouples)
                {

                    mmCount += c.MismatchesBoth;
                    leCount += c.LengthBoth;
                }
                //Calculate average length
                double averageLengthAssoc = (double)leCount / (2.0 * (double)maCount);

                //Calculate standard deviation
                double squareSum = 0.0;
                foreach (SSRCouple c in activeCouples)
                {
                    squareSum += Math.Pow((c.First.Length - averageLengthAssoc), 2.0);
                    squareSum += Math.Pow((c.Second.Length - averageLengthAssoc), 2.0);
                }
                double avStdDev = Math.Sqrt(squareSum / (maCount * 2.0));




                //Text for all motif Associations and all couples
                sb.Append("MOTIF ASSOCIATION STATISTIC\n");
                sb.Append("All SSR-Couples\tHeterocouples\tHomocouples\n");
                sb.Append(String.Format("{0}\t{1}\t{2}\n\n", all_couples.Count, heteroCouples.Count, homoCouples.Count));

                Debug.WriteLine(String.Format("{0} SSR-statistics module; Display statistics; Displayed overall SSR-couple statistics", DateTime.Now.TimeOfDay));




                //Text for all motif Associations and active couples
                sb.Append(motifAssocText);
                sb.Append(String.Format("SSR-Couples considered for statistic {0,6}  - SSR-Couple/1000 SSRs{1,7:f}\n",
                    maCount, (1000.0 * ((double)maCount) / (this.subsetSSRs == null ? 0 : this.subsetSSRs.Count))
                    ));
                sb.Append("Average_Mismatches/SSR_in_SSR-Couple\tAverage_Length/SSR_in_SSR-Couple\t Standard_Deviation_Length/SSR_in_SSR-Couple\t SSR-Couple/Mbp\n");
                sb.Append(String.Format("{0,6:f}\t{1,6:f}\t{2,6:f}\t{3,6:f} \n\n",
                     (((double)mmCount) / (2.0 * (double)maCount)), averageLengthAssoc, avStdDev, 1000000.0 / ((double)nucleotideSummaryInfo.nucleotideCount / (double)maCount)));

                //TYPE 1
                if (parameters.motifAssoc_FS)
                {

                    sb.Append(String.Format("Motif associations, Fully standardized (See Manual):\n"));
                    sb.Append(motifAssocText);


                    List<SSRStatisticMotifAssociation> ma1 = couStat.GetMotifAssociationStatistics(MotifAssociationType.FullyStandardized);
                    sb.Append(SSRStatisticMotifAssociation.ToStringHeader(detailLevel) + '\n');
                    foreach (SSRStatisticMotifAssociation ma in ma1)
                    {
                        if (ma.Count >= minCount)
                        {
                            sb.Append(ma.ToString(detailLevel) + '\n');
                        }
                    }
                    sb.Append('\n');
                    Debug.WriteLine(String.Format("{0} SSR-statistics module; Display statistics; Displayed fully standardized motif associations", DateTime.Now.TimeOfDay));

                }


                //TYPE 2
                if (parameters.motifAssoc_CS)
                {

                    sb.Append(String.Format("Motif associations, Conformation standardized (See Manual):\n"));
                    sb.Append(motifAssocText);

                    List<SSRStatisticMotifAssociation> ma2 = couStat.GetMotifAssociationStatistics(MotifAssociationType.ConformationStandardized);
                    sb.Append(SSRStatisticMotifAssociation.ToStringHeader(detailLevel) + '\n');
                    foreach (SSRStatisticMotifAssociation ma in ma2)
                    {
                        if (ma.Count >= minCount)
                        {
                            sb.Append(ma.ToString(detailLevel) + '\n');
                        }
                    }
                    sb.Append('\n');
                    Debug.WriteLine(String.Format("{0} SSR-statistics module; Display statistics; Displayed conformation standardized motif associations", DateTime.Now.TimeOfDay));
                }


                //TYPE 3
                if (parameters.motifAssoc_PSB)
                {
                    sb.Append(String.Format("Motif associations, Partially standardized both strands (See Manual):\n"));
                    sb.Append(motifAssocText);


                    List<SSRStatisticMotifAssociation> ma3 = couStat.GetMotifAssociationStatistics(MotifAssociationType.PartiallyStandardizedBothStrands);
                    sb.Append(SSRStatisticMotifAssociation.ToStringHeader(detailLevel) + '\n');
                    foreach (SSRStatisticMotifAssociation ma in ma3)
                    {
                        if (ma.Count >= minCount)
                        {
                            sb.Append(ma.ToString(detailLevel) + '\n');
                        }
                    }
                    sb.Append('\n');
                    Debug.WriteLine(String.Format("{0} SSR-statistics module; Display statistics; Displayed partially standardized both strands motif associations", DateTime.Now.TimeOfDay));
                }


                //TYPE 4
                if (parameters.motifAssoc_PSS)
                {
                    sb.Append(String.Format("Motif associations, Partially standardized (See Manual):\n"));
                    sb.Append(motifAssocText);


                    List<SSRStatisticMotifAssociation> ma4 = couStat.GetMotifAssociationStatistics(MotifAssociationType.PartiallyStandardized);
                    sb.Append(SSRStatisticMotifAssociation.ToStringHeader(detailLevel) + '\n');
                    foreach (SSRStatisticMotifAssociation ma in ma4)
                    {
                        if (ma.Count >= minCount)
                        {
                            sb.Append(ma.ToString(detailLevel) + '\n');
                        }
                    }
                    sb.Append('\n');
                    Debug.WriteLine(String.Format("{0} SSR-statistics module; Display statistics; Displayed partially standardized single strand motif associations", DateTime.Now.TimeOfDay));
                }


            }
#if DEBUG
            stopw.Stop();
            Debug.WriteLine(String.Format("{0} SciRoKo SSR-statistics model; Displayed SSR statistics in {1} ms", DateTime.Now.TimeOfDay, stopw.ElapsedMilliseconds));
            Debug.WriteLine("");
#endif
            return sb;


        }





    }



    /// <summary>
    /// Regulates the Access to the ini file
    /// </summary>
    public class Model_IniFileAccess
    {
        private XmlDocument settings;
        private string startupPath;
        private static class SettingsRetrieval
        {
            public static string startupPath = @"/Settings/StartUpPath";
        }

        public Model_IniFileAccess()
        {
            LoadIni();
        }

        public void LoadIni()
        {
            string path = Path.GetDirectoryName(System.Reflection.Assembly.GetEntryAssembly().Location);
            path = Path.Combine(path, "initialize.ini");
            XmlTextReader tr = new XmlTextReader(path);
            settings = new XmlDocument();


            try
            {
                settings.Load(tr);
                startupPath = settings.SelectSingleNode(SettingsRetrieval.startupPath).InnerText;

            }
            catch 
            {

                MessageBox.Show("Initialize.ini file could not be found. Creating initialize.ini");
                startupPath = Path.GetDirectoryName(System.Reflection.Assembly.GetEntryAssembly().Location);
                CreateIni(startupPath);
                LoadIni();
            }
            finally
            {

                tr.Close();
            }


        }
        /// <summary>
        /// Returns the startup Path for the File-Dialogs
        /// </summary>
        public string StartupPath
        {
            get
            {
                return startupPath;
            }
            set
            {
                SavePath(value);
            }
        }
        private void SavePath(string path)
        {


            settings.SelectSingleNode(SettingsRetrieval.startupPath).InnerText = path;

            string saveSettingsPath = Path.GetDirectoryName(System.Reflection.Assembly.GetEntryAssembly().Location);
            saveSettingsPath = Path.Combine(saveSettingsPath, "initialize.ini");
            System.Xml.XmlTextWriter xmlText = new XmlTextWriter(saveSettingsPath, System.Text.Encoding.ASCII);
            try
            {
                startupPath = path;


                xmlText.Formatting = System.Xml.Formatting.Indented;
                xmlText.Indentation = 1;
                xmlText.IndentChar = ' ';
                settings.Save(xmlText);
            }
            finally
            {
                xmlText.Close();
            }
        }

        public void CreateIni(string startupPathParam)
        {
            System.Xml.XmlTextWriter xmlText = new XmlTextWriter(Path.Combine(Path.GetDirectoryName(System.Reflection.Assembly.GetEntryAssembly().Location), "initialize.ini"), System.Text.Encoding.ASCII);
            xmlText.Formatting = System.Xml.Formatting.Indented;
            xmlText.Indentation = 2; //Drei Einrcken
            xmlText.IndentChar = ' '; //Einrckcharakter ist die Space taste

            //Start first document
            xmlText.WriteStartDocument(true);
            xmlText.WriteComment("Startup settings for SciRoKo");

            //Create namespace
            xmlText.WriteStartElement("Settings");

            xmlText.WriteElementString("StartUpPath", startupPathParam);


            //close first documentt
            try
            {
                xmlText.WriteEndDocument();
            }
            finally
            {
                xmlText.Close();
            }
        }

    }

    public class Model_LittleHelper
    {

        private List<IPositionable> toFlank=null;
        private string outputFile=null;
        private string[] inputFiles=null;

        public Model_LittleHelper()
        {
        }

        public void ExtractFlankingSequences(ExtractFlankingSettings settings)
        {
            if (toFlank == null || outputFile == null || inputFiles == null) throw new ArgumentNullException("Error during extraction process; Not all necessary variables have been instantiated");

#if DEBUG
            System.Diagnostics.Stopwatch stopw= new Stopwatch();
            int count = 0;
            stopw.Start();
            
#endif
            ISequenceReader<NucleotideSequence> reader;
            Bio.IO.Seq.Misc.FastaMultiFileReader_Sequence<NucleotideSequence> fr = new Bio.IO.Seq.Misc.FastaMultiFileReader_Sequence<NucleotideSequence>(inputFiles,new DNASequenceValidator_ATCGNYRWSKMBDHV(),false);
            if (settings.totalFlankMode) reader = new SequenceReaderDecorator_ExtractFlankingSequences<NucleotideSequence>(fr, toFlank,true, settings.totalFlankLength);
            else reader = new SequenceReaderDecorator_ExtractFlankingSequences<NucleotideSequence>(fr, toFlank,true, settings.flank5Prime, settings.flank3Prime);
            StreamWriter sw = new StreamWriter(outputFile,false, Encoding.ASCII);
            FastaWriter<NucleotideSequence> writer = new FastaWriter<NucleotideSequence>(sw);

            List<NucleotideSequence> temp;
            while ((temp = reader.GetSequences()).Count > 0)
            {
                writer.WriteSequences(temp);
#if DEBUG
                count+=temp.Count;
#endif
            }
            sw.Close();

#if DEBUG
            stopw.Stop();
            System.Diagnostics.Debug.WriteLine(String.Format("{0}: SciRoKo Model Little Helper; Extracted flanking sequences for {1} sequences into {2} files in {3} ms", DateTime.Now.TimeOfDay,toFlank.Count,count,stopw.ElapsedMilliseconds));
#endif
            reader = null;
            fr = null;
            System.GC.Collect();

        }

        /// <summary>
        /// Set the sequences which should be flanked
        /// </summary>
        public List<IPositionable> SequencesToFlank
        {
            set
            {
                this.toFlank = value;
            }
        }

        public string SaveFileForFlankingSequences
        {
            set
            {
                this.outputFile = value;
            }
        }

        public string[] FilesContainingTheFlanks
        {
            set
            {
                this.inputFiles = value;
            }
        }



        public void DigestSequences(string[] sequenceFiles, DigestionSettings settings)
        {
            string[] sequencesToPretreat = sequenceFiles;
            string folderToSavePretreatet = Path.GetDirectoryName(sequencesToPretreat[0]);
            folderToSavePretreatet += "\\Digested";
            if (!Directory.Exists(folderToSavePretreatet)) Directory.CreateDirectory(folderToSavePretreatet);
            folderToSavePretreatet += "\\";
            string[] pretreatSequences = new string[sequencesToPretreat.Length];

            for (int i = 0; i < sequencesToPretreat.Length; i++)
            {
                pretreatSequences[i] = folderToSavePretreatet + Path.GetFileName(sequencesToPretreat[i]);
            }

            for (int i = 0; i < sequencesToPretreat.Length; i++)
            {
                StreamReader sr = new StreamReader(sequencesToPretreat[i], Encoding.ASCII);
                StreamWriter sw = new StreamWriter(pretreatSequences[i], false, Encoding.ASCII);

                List<NucleotideSequence> sequences;


                ISequenceReader<NucleotideSequence> seqr = new FastaReaderSequenceChopper<NucleotideSequence>(sr, new DNASequenceValidator_ATCGNYRWSKMBDHV(), true, settings.digestToSize, settings.overhead);
                seqr = new SequenceReaderDecorator_UniqueSeqID<NucleotideSequence>(seqr);
                ISequenceWriter<NucleotideSequence> seqwr = new FastaWriter<NucleotideSequence>(sw);
                try
                {
                    while ((sequences = seqr.GetSequences()).Count > 0)
                    {
                        seqwr.WriteSequences(sequences);

                        sequences.Clear();
                    }
                }
                finally
                {
                    sw.Close();
                    sr.Close();
                }

            }
        }
    }


    #region SciRoKo_MainStates
    public interface ISciRoKoMain_State
    {
        void UpdateView();
    }
    /// <summary>
    /// Representing inner state of controll element, updating view, no state specified, state pattern
    /// </summary>
    public class SciRoKoState_None : ISciRoKoMain_State
    {
        private ISciRoKoView view;

        public SciRoKoState_None(ISciRoKoView view)
        {
            this.view = view;
            UpdateView();
        }

        public void UpdateView()
        {
            view.Button_DisplayStatisticsEnabled = false;
            view.Button_ExportSSRsEnabled = false;
            view.Button_LoadSequencesEnabled = true;
            view.Button_LoadSSRsEnabled = true;
            view.Button_SaveSearchResultsEnabled = true;
            view.Button_StartSearchEnabled = false;
            view.Button_ExtractFlankingSequenceEnabled = false;
            view.RichTextBox_AnalyzedText = "";
            view.RichTextBox_FileStatusText = "";

            //Statistics
            view.TextBox_MotifAssociationMaxDistanceEnabled = true;
            view.GroupBox_MotifStatisticsEnabled = true;
            view.CheckBox_ShowInfoEnabled = true;
            view.CheckBox_ShowLengthInfoEnabled = true;
            view.GroupBox_SearchResultSortingEnabled = true;
            view.TextBox_MinimumRequiredScoreEnabled = true;
            view.TextBox_LowerBoundaryEnabled = true;
            view.TextBox_UpperBoundaryEnabled = true;
            view.RadioButton_DisplaySearchResultsEnabled = true;
            view.RadioButton_DisplayStatisticsEnabled = true;



        }

    }


    /// <summary>
    /// Representing inner state of controll element, updating view, sequences for SSR-search have been chosen, state pattern
    /// </summary>
    public class SciRoKoState_SequencesChosen : ISciRoKoMain_State
    {
        ISciRoKoView view;
        Model_SSRSearch mSearch;
        public SciRoKoState_SequencesChosen(ISciRoKoView view, Model_SSRSearch mSearch)
        {
            this.view = view;
            this.mSearch = mSearch;
            UpdateView();
        }

        public void UpdateView()
        {
            view.Button_DisplayStatisticsEnabled = false;
            view.Button_ExportSSRsEnabled = false;
            view.Button_LoadSequencesEnabled = false;
            view.Button_LoadSSRsEnabled = false;
            view.Button_SaveSearchResultsEnabled = true;
            view.Button_StartSearchEnabled = false;
            view.Button_ExtractFlankingSequenceEnabled = false;

            //Statistics
            view.TextBox_MotifAssociationMaxDistanceEnabled = true;
            view.GroupBox_MotifStatisticsEnabled = true;
            view.CheckBox_ShowInfoEnabled = true;
            view.CheckBox_ShowLengthInfoEnabled = true;
            view.GroupBox_SearchResultSortingEnabled = true;
            view.TextBox_MinimumRequiredScoreEnabled = true;
            view.TextBox_LowerBoundaryEnabled = true;
            view.TextBox_UpperBoundaryEnabled = true;
            view.RadioButton_DisplaySearchResultsEnabled = true;
            view.RadioButton_DisplayStatisticsEnabled = true;

            //Prepare the string for the form
            StringBuilder sb = new StringBuilder();
            string[] seq = mSearch.SequenceFiles;
            sb.Append(String.Format("SSR Search Mode:\n\nInput files have been choosen.\n {0,2} Files have been selected for analysis\n", seq.GetLength(0)));
            foreach (string s in seq)
            {
                sb.Append(Path.GetFileName(s) + '\n');
            }
            sb.Append("\nChoose the Output-file before you can start the SSR-Search!!\n");
            sb.Append("If you want to select other files or enter the statistic mode hit the Reset-Button");


            view.RichTextBox_AnalyzedText = "";
            view.RichTextBox_FileStatusText = sb.ToString();

        }

    }

    public class SciRoKoState_SaveFileChosen : ISciRoKoMain_State
    {
        private ISciRoKoView view;
        private Model_SSRSearch mSearch;
        public SciRoKoState_SaveFileChosen(ISciRoKoView view, Model_SSRSearch mSearch)
        {
            this.view = view;
            this.mSearch = mSearch;
            UpdateView();
        }

        public void UpdateView()
        {

            view.Button_DisplayStatisticsEnabled = false;
            view.Button_ExportSSRsEnabled = false;
            view.Button_LoadSequencesEnabled = true;
            view.Button_LoadSSRsEnabled = false;
            view.Button_SaveSearchResultsEnabled = false;
            view.Button_StartSearchEnabled = false;
            view.Button_ExtractFlankingSequenceEnabled = false;

            //Statistics
            view.TextBox_MotifAssociationMaxDistanceEnabled = true;
            view.GroupBox_MotifStatisticsEnabled = true;
            view.CheckBox_ShowInfoEnabled = true;
            view.CheckBox_ShowLengthInfoEnabled = true;
            view.GroupBox_SearchResultSortingEnabled = true;
            view.TextBox_MinimumRequiredScoreEnabled = true;
            view.TextBox_LowerBoundaryEnabled = true;
            view.TextBox_UpperBoundaryEnabled = true;
            view.RadioButton_DisplaySearchResultsEnabled = true;
            view.RadioButton_DisplayStatisticsEnabled = true;

            StringBuilder sb = new StringBuilder();
            sb.Append(String.Format("SSR Search Mode:\n\nOutput file has been chosen.\n"));
            sb.Append(String.Format("{0} \n", Path.GetFileName(mSearch.SaveFile)));
            sb.Append("\nChoose the files containing the sequences you want to analyise before you can start the SSR-Search!!\n");
            sb.Append("If you want to select other files or enter the statistic mode hit the Reset-Button");


            view.RichTextBox_AnalyzedText = "";
            view.RichTextBox_FileStatusText = sb.ToString();




        }

    }


    public class SciRoKoState_SequencesAndSaveFilesChosen : ISciRoKoMain_State
    {
        private ISciRoKoView view;
        private Model_SSRSearch mSearch;
        public SciRoKoState_SequencesAndSaveFilesChosen(ISciRoKoView view, Model_SSRSearch mSearch)
        {
            this.view = view;
            this.mSearch = mSearch;
            UpdateView();
        }

        public void UpdateView()
        {

            view.Button_DisplayStatisticsEnabled = false;
            view.Button_ExportSSRsEnabled = false;
            view.Button_LoadSequencesEnabled = false;
            view.Button_LoadSSRsEnabled = false;
            view.Button_SaveSearchResultsEnabled = false;
            view.Button_StartSearchEnabled = true;
            view.Button_ExtractFlankingSequenceEnabled = false;

            //Statistics
            view.TextBox_MotifAssociationMaxDistanceEnabled = true;
            view.GroupBox_MotifStatisticsEnabled = true;
            view.CheckBox_ShowInfoEnabled = true;
            view.CheckBox_ShowLengthInfoEnabled = true;
            view.GroupBox_SearchResultSortingEnabled = true;
            view.TextBox_MinimumRequiredScoreEnabled = true;
            view.TextBox_LowerBoundaryEnabled = true;
            view.TextBox_UpperBoundaryEnabled = true;
            view.RadioButton_DisplaySearchResultsEnabled = true;
            view.RadioButton_DisplayStatisticsEnabled = true;

            string[] seq = mSearch.SequenceFiles;

            StringBuilder sb = new StringBuilder();
            sb.Append("SSR Search Mode:\n\n");
            sb.Append(String.Format("Input and Output files have been chosen\n Input files: {0,2}\n", seq.GetLength(0)));
            foreach (string s in seq)
            {
                sb.Append(Path.GetFileName(s) + '\n');
            }
            sb.Append("\nOutput file chosen:\n");
            sb.Append(Path.GetFileName(mSearch.SaveFile) + '\n');
            sb.Append("\nPress the Start search Button to beginn the SSR-search!!\n");
            sb.Append("If you want to select other files or enter the statistic mode hit the Reset-Button");

            view.RichTextBox_AnalyzedText = "";
            view.RichTextBox_FileStatusText = sb.ToString();
        }

    }

    public class SciRoKoState_SearchFinished : ISciRoKoMain_State
    {

        private ISciRoKoView view;
        private Model_SSRSearch mSearch;
        public SciRoKoState_SearchFinished(ISciRoKoView view, Model_SSRSearch mSearch)
        {
            this.view = view;
            this.mSearch = mSearch;
            UpdateView();
        }

        public void UpdateView()
        {
            view.Button_DisplayStatisticsEnabled = false;
            view.Button_ExportSSRsEnabled = false;
            view.Button_LoadSequencesEnabled = false;
            view.Button_LoadSSRsEnabled = true;
            view.Button_SaveSearchResultsEnabled = true;
            view.Button_StartSearchEnabled = false;
            view.Button_ExtractFlankingSequenceEnabled = false;

            //Statistics
            view.TextBox_MotifAssociationMaxDistanceEnabled = true;
            view.GroupBox_MotifStatisticsEnabled = true;
            view.CheckBox_ShowInfoEnabled = true;
            view.CheckBox_ShowLengthInfoEnabled = true;
            view.GroupBox_SearchResultSortingEnabled = true;
            view.TextBox_MinimumRequiredScoreEnabled = true;
            view.TextBox_LowerBoundaryEnabled = true;
            view.TextBox_UpperBoundaryEnabled = true;
            view.RadioButton_DisplaySearchResultsEnabled = true;
            view.RadioButton_DisplayStatisticsEnabled = true;


            StringBuilder sb = new StringBuilder();
            string[] seq = mSearch.SequenceFiles;
            sb.Append(string.Format("Search has been finished in {0,5:f} seconds\n\n", mSearch.SearchTime.TotalSeconds));
            sb.Append(String.Format("SSR Search Mode:\n\nInput files remain chosen.\n {0,2} file(s) have been selected for analysis\n", seq.GetLength(0)));
            foreach (string s in seq)
            {
                sb.Append(Path.GetFileName(s) + '\n');
            }
            sb.Append("\nChoose another Output-file to search the same sequences again, or load the output file into the statistic-mode\n");



            view.RichTextBox_AnalyzedText = "";
            view.RichTextBox_FileStatusText = sb.ToString();

        }

    }

    public class SciRoKoState_SSRLoaded : ISciRoKoMain_State
    {
        private ISciRoKoView view;
        private Model_SSRStatistic mStatistic;
        public SciRoKoState_SSRLoaded(ISciRoKoView view, Model_SSRStatistic mStatistic)
        {
            this.view = view;
            this.mStatistic = mStatistic;
            UpdateView();
        }

        public void UpdateView()
        {
            view.Button_DisplayStatisticsEnabled = true;
            view.Button_ExportSSRsEnabled = true;
            view.Button_LoadSequencesEnabled = false;
            view.Button_LoadSSRsEnabled = true;
            view.Button_SaveSearchResultsEnabled = false;
            view.Button_StartSearchEnabled = false;
            view.Button_ExtractFlankingSequenceEnabled = true;

            //Statistics
            view.TextBox_MotifAssociationMaxDistanceEnabled = true;
            view.GroupBox_MotifStatisticsEnabled = true;
            view.CheckBox_ShowInfoEnabled = true;
            view.CheckBox_ShowLengthInfoEnabled = true;
            view.GroupBox_SearchResultSortingEnabled = true;
            view.TextBox_MinimumRequiredScoreEnabled = true;
            view.TextBox_LowerBoundaryEnabled = true;
            view.TextBox_UpperBoundaryEnabled = true;
            view.RadioButton_DisplaySearchResultsEnabled = true;
            view.RadioButton_DisplayStatisticsEnabled = true;

            StringBuilder sb = new StringBuilder();
            string[] msa = mStatistic.MicrosatelliteFiles;
            sb.Append("SSR Statistic Mode - SSRs loaded:\n\n" + String.Format("Input files have been chosen.\n {0,2} file(s) have been selected for analysis\n", msa.GetLength(0)));
            sb.Append(String.Format("Format of the input files: {0}\n", SSRUtility.GetSSRFileModeFromExtension(Path.GetExtension(msa[0])).ToString()));
            foreach (string s in msa)
            {
                sb.Append(Path.GetFileName(s) + '\n');
            }
            sb.Append(String.Format("\nThe input files contain in total: {0} SSRs, the selected subset: {1}\n", mStatistic.SSRTotalCount, mStatistic.SSRSubsetCount));
            sb.Append(String.Format("Number of total nucleotids: {0} - Number of total fastas: {1}\n", mStatistic.SequenceSummaryInfo.nucleotideCount, mStatistic.SequenceSummaryInfo.fileCount));
            sb.Append("\nYou can either export the selected SSR-subset or display the statistics of the subset!!\n");
            sb.Append("If you want to enter the search-mode hit the Reset-Button");


            view.RichTextBox_AnalyzedText = "";
            view.RichTextBox_FileStatusText = sb.ToString();

        }

    }

    public class SciRoKoState_SSRCouples : ISciRoKoMain_State
    {
        private ISciRoKoView view;
        private Model_SSRStatistic mStatistic;

        public SciRoKoState_SSRCouples(ISciRoKoView view, Model_SSRStatistic mStatistic)
        {
            this.view = view;
            this.mStatistic = mStatistic;
            UpdateView();
        }

        public void UpdateView()
        {
            view.Button_DisplayStatisticsEnabled = true;
            view.Button_ExportSSRsEnabled = false;
            view.Button_LoadSequencesEnabled = false;
            view.Button_LoadSSRsEnabled = true;
            view.Button_SaveSearchResultsEnabled = false;
            view.Button_StartSearchEnabled = false;
            view.Button_ExtractFlankingSequenceEnabled = true;

            //Statistics
            view.TextBox_MotifAssociationMaxDistanceEnabled = false;
            view.GroupBox_MotifStatisticsEnabled = false;
            view.CheckBox_ShowInfoEnabled = false;
            view.CheckBox_ShowLengthInfoEnabled = false;
            view.GroupBox_SearchResultSortingEnabled = false;
            view.TextBox_MinimumRequiredScoreEnabled = false;
            view.TextBox_LowerBoundaryEnabled = false;
            view.TextBox_UpperBoundaryEnabled = false;
            view.RadioButton_DisplayStatitsticsChecked = true;

            view.RadioButton_DisplaySearchResultsEnabled = false;
            view.RadioButton_DisplayStatisticsEnabled = false;
            string[] msa = mStatistic.MicrosatelliteFiles;


            StringBuilder sb = new StringBuilder();
            sb.Append(String.Format("SSR Statistic Mode - SSR couples loaded:\n\nInput files have been chosen.\n {0,2} file(s) have been selected for analysis\n", msa.GetLength(0)));
            foreach (string s in msa)
            {
                sb.Append(Path.GetFileName(s) + '\n');
            }
            sb.Append(String.Format("\nThe input files contain in total: {0} SSR couples\n", mStatistic.SSRCoupleCount));
            sb.Append(String.Format("Number of total nucleotids: {0} - Number of total fastas: {1}\n", mStatistic.SequenceSummaryInfo.nucleotideCount, mStatistic.SequenceSummaryInfo.fileCount));
            sb.Append("\nYou can either export the selected SSR-subset or display the statistics of the subset!!\n");
            sb.Append("If you want to enter the search-mode hit the Reset-Button");


            view.RichTextBox_AnalyzedText = "";
            view.RichTextBox_FileStatusText = sb.ToString();


        }

    }

    public class SciRoKoState_Deactivated : ISciRoKoMain_State
    {
        private ISciRoKoView view;
        public SciRoKoState_Deactivated(ISciRoKoView view)
        {
            this.view = view;
            UpdateView();
        }

        public void UpdateView()
        {
            view.Button_DisplayStatisticsEnabled = false;
            view.Button_ExportSSRsEnabled = true;
            view.Button_LoadSequencesEnabled = false;
            view.Button_LoadSSRsEnabled = false;
            view.Button_SaveSearchResultsEnabled = false;
            view.Button_StartSearchEnabled = false;
            view.Button_ExtractFlankingSequenceEnabled = false;

            //Statistics
            view.TextBox_MotifAssociationMaxDistanceEnabled = false;
            view.GroupBox_MotifStatisticsEnabled = false;
            view.CheckBox_ShowInfoEnabled = false;
            view.CheckBox_ShowLengthInfoEnabled = false;
            view.GroupBox_SearchResultSortingEnabled = false;
            view.TextBox_MinimumRequiredScoreEnabled = false;
            view.TextBox_LowerBoundaryEnabled = false;
            view.TextBox_UpperBoundaryEnabled = false;
            view.RadioButton_DisplaySearchResultsEnabled = false;
            view.RadioButton_DisplayStatisticsEnabled = false;



            StringBuilder sb = new StringBuilder();
            sb.Append(String.Format("SciRoKo is deactivated; Please obtain a proper license"));



            view.RichTextBox_AnalyzedText = "";
            view.RichTextBox_FileStatusText = sb.ToString();
        }

    }
    #endregion



    public struct SSRSubsetBoundary
    {
        public int upperBoundary;
        public int lowerBoundary;
        public int minimumScore;
    }
    public struct SSRStatisticParameters
    {
        public int minimumFrequencyInCount;
        public int maxDistanceForAssociation;
        public bool displayMotifLengthStatistics;
        public bool displayMotifStatistics;
        public bool displayMotifAssociationStatistics;
        public bool motifStat_FS;
        public bool motifStat_PS;
        public bool motifAssoc_FS;
        public bool motifAssoc_CS;
        public bool motifAssoc_PSB;
        public bool motifAssoc_PSS;
        public SSR.SSRSort.SSRSortMode ssrSortMode;
        public CoupleSubtype couplesSubtype;
        public Bio.Repetitive.Microsat.Statistics.SSRStatistics_DetailLevel statisticDetailLevel;




        public enum CoupleSubtype
        {
            Homocouples,
            Heterocouples,
            All
        }

    }


    public struct DigestionSettings
    {
        public int digestToSize;
        public int overhead;
    }


    public struct ExtractFlankingSettings
    {
        public bool totalFlankMode;
        public List<IPositionable> toFlank;
        public int flank5Prime;
        public int flank3Prime;
        public int totalFlankLength;
    }



}